# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, Union, Optional

from .._models import BaseModel

__all__ = ["WorkflowItem"]


class WorkflowItem(BaseModel):
    workflow_name: str

    workflow_alias: Optional[str] = None

    workflow_inputs: Optional[Dict[str, Union[str, Dict[str, Union[str, object]]]]] = None



"""
Tests for error handling improvements in streaming module.

These tests verify that the streaming code gracefully handles:
- Invalid JSON in SSE data
- Malformed UTF-8 sequences
- RemoteProtocolError logging at INFO level
"""
# pyright: reportUnknownVariableType=false, reportUnknownMemberType=false, reportUnknownArgumentType=false
import logging
from typing import AsyncIterator
from unittest.mock import Mock, PropertyMock

import httpx
import pytest

from sgp_agents._streaming import SSEDecoder, AsyncStream, ServerSentEvent


def create_mock_client() -> Mock:
    """Create a mock client with a real SSEDecoder."""
    mock_client = Mock()
    mock_client._make_sse_decoder.return_value = SSEDecoder()
    return mock_client


class TestJSONErrorHandling:
    """Test that invalid JSON in SSE data is handled gracefully."""

    def test_json_decode_error_raises_valueerror(self) -> None:
        """Verify that invalid JSON raises ValueError with clear message."""
        sse = ServerSentEvent(
            event=None,
            data="this is not valid json {",
            id=None,
            retry=None
        )

        with pytest.raises(ValueError) as exc_info:
            sse.json()

        assert "invalid JSON" in str(exc_info.value)
        assert "Expecting" in str(exc_info.value)  # From JSONDecodeError

    def test_json_decode_error_logs_at_error_level(self, caplog: pytest.LogCaptureFixture) -> None:
        """Verify that invalid JSON is logged at ERROR level."""
        sse = ServerSentEvent(
            event=None,
            data='{"incomplete": ',
            id=None,
            retry=None
        )

        with caplog.at_level(logging.ERROR):
            try:
                sse.json()
            except ValueError:
                pass  # Expected

        # Check that error was logged
        assert any("[SSE] Invalid JSON" in record.message for record in caplog.records)

    def test_valid_json_still_works(self) -> None:
        """Verify that valid JSON parsing still works correctly."""
        sse = ServerSentEvent(
            event=None,
            data='{"key": "value", "number": 42}',
            id=None,
            retry=None
        )

        result = sse.json()
        assert result == {"key": "value", "number": 42}

    def test_json_error_includes_data_preview(self, caplog: pytest.LogCaptureFixture) -> None:
        """Verify that error log includes data preview for debugging."""
        long_data = "x" * 500 + " invalid json"
        sse = ServerSentEvent(
            event=None,
            data=long_data,
            id=None,
            retry=None
        )

        with caplog.at_level(logging.ERROR):
            try:
                sse.json()
            except ValueError:
                pass

        # Should truncate at 200 chars
        error_message = caplog.records[0].message
        assert "Data preview:" in error_message
        # Make sure it doesn't include the full 500+ char string
        assert len(error_message) < 400


class TestUTF8ErrorHandling:
    """Test that malformed UTF-8 sequences are handled gracefully."""

    @pytest.mark.asyncio
    async def test_utf8_decode_error_skips_line(self, caplog: pytest.LogCaptureFixture) -> None:
        """Verify that invalid UTF-8 line is skipped with warning."""
        mock_response = Mock()

        async def chunks_with_invalid_utf8() -> AsyncIterator[bytes]:
            # Valid SSE event with invalid UTF-8 in one line
            yield b"data: valid line\n"
            yield b"data: \xff\xfe invalid UTF-8\n"  # Invalid UTF-8 sequence
            yield b"data: another valid line\n\n"

        mock_response.aiter_bytes = chunks_with_invalid_utf8
        type(mock_response)._request = PropertyMock(return_value=None)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=dict,
            response=mock_response,
            client=mock_client,
        )

        with caplog.at_level(logging.WARNING):
            events = []
            async for event in stream._iter_events():
                events.append(event)

        # Should receive 1 event (the invalid line was skipped)
        assert len(events) == 1
        assert "valid line" in events[0].data
        assert "another valid line" in events[0].data

        # Check that warning was logged
        assert any("Failed to decode line as UTF-8" in record.message for record in caplog.records)

    @pytest.mark.asyncio
    async def test_utf8_error_includes_hex_preview(self, caplog: pytest.LogCaptureFixture) -> None:
        """Verify that UTF-8 error log includes hex preview for debugging."""
        mock_response = Mock()

        async def chunks_with_invalid_utf8() -> AsyncIterator[bytes]:
            yield b"data: \xff\xfe\xfd\n\n"  # Invalid UTF-8

        mock_response.aiter_bytes = chunks_with_invalid_utf8
        type(mock_response)._request = PropertyMock(return_value=None)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=dict,
            response=mock_response,
            client=mock_client,
        )

        with caplog.at_level(logging.WARNING):
            events = []
            async for event in stream._iter_events():
                events.append(event)

        # Check that hex preview is in log
        warning_message = caplog.records[0].message
        assert "hex" in warning_message.lower()
        assert "fffefd" in warning_message  # Hex representation of the invalid bytes

    def test_sync_utf8_decode_error_skips_line(self, caplog: pytest.LogCaptureFixture) -> None:
        """Verify that sync decoder also skips invalid UTF-8."""
        from typing import Iterator
        decoder = SSEDecoder()

        def chunks_with_invalid_utf8() -> Iterator[bytes]:
            yield b"data: valid\n\n"
            yield b"data: \xff\xfe\n\n"  # Invalid UTF-8
            yield b"data: also valid\n\n"

        with caplog.at_level(logging.WARNING):
            events = list(decoder.iter_bytes(chunks_with_invalid_utf8()))

        # Should receive 2 events (invalid was skipped)
        assert len(events) == 2
        assert events[0].data == "valid"
        assert events[1].data == "also valid"

        # Check warning was logged
        assert any("Failed to decode line as UTF-8" in record.message for record in caplog.records)


class TestRemoteProtocolErrorLogging:
    """Test that RemoteProtocolError is logged at INFO, not ERROR."""

    @pytest.mark.asyncio
    async def test_remote_protocol_error_logs_at_info_level(self, caplog: pytest.LogCaptureFixture) -> None:
        """Verify RemoteProtocolError is logged at INFO level in decoder."""
        mock_response = Mock()

        async def chunks_then_error() -> AsyncIterator[bytes]:
            yield b"data: chunk0\n\n"
            raise httpx.RemoteProtocolError("Server closed connection")

        mock_response.aiter_bytes = chunks_then_error
        type(mock_response)._request = PropertyMock(return_value=None)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=dict,
            response=mock_response,
            client=mock_client,
        )

        with caplog.at_level(logging.INFO):
            events = []
            try:
                async for event in stream._iter_events():
                    events.append(event)
            except httpx.RemoteProtocolError:
                pass  # Expected

        # Check that it was logged at INFO, not ERROR
        info_logs = [r for r in caplog.records if r.levelname == "INFO"]
        error_logs = [r for r in caplog.records if r.levelname == "ERROR"]

        assert len(events) == 1
        assert any("Server closed connection" in r.message or "Stream ended" in r.message for r in info_logs)
        # Should NOT be in error logs
        assert not any("RemoteProtocolError" in r.message for r in error_logs)

    @pytest.mark.asyncio
    async def test_remote_protocol_error_by_class_name(self, caplog: pytest.LogCaptureFixture) -> None:
        """Verify class name check catches RemoteProtocolError."""
        # Create a mock exception with the same name but different instance
        class FakeRemoteProtocolError(Exception):
            """Fake exception with same name to test class name matching."""
            pass

        # Monkey-patch the class name
        FakeRemoteProtocolError.__name__ = "RemoteProtocolError"

        mock_response = Mock()

        async def chunks_then_error() -> AsyncIterator[bytes]:
            yield b"data: chunk0\n\n"
            raise FakeRemoteProtocolError("Fake error")

        mock_response.aiter_bytes = chunks_then_error
        type(mock_response)._request = PropertyMock(return_value=None)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=dict,
            response=mock_response,
            client=mock_client,
        )

        with caplog.at_level(logging.INFO):
            try:
                async for _ in stream._iter_events():
                    pass
            except FakeRemoteProtocolError:
                pass  # Expected

        # Should be logged at INFO because class name matches
        info_logs = [r for r in caplog.records if r.levelname == "INFO"]
        assert any("Stream ended" in r.message for r in info_logs)


class TestIntegrationScenarios:
    """Test complete scenarios combining multiple error types."""

    @pytest.mark.asyncio
    async def test_stream_continues_after_invalid_utf8_line(self) -> None:
        """Verify stream continues processing after encountering invalid UTF-8."""
        mock_response = Mock()

        async def mixed_chunks() -> AsyncIterator[bytes]:
            yield b"data: {\"valid\": 1}\n\n"
            yield b"data: \xff\xfe\n"  # Invalid UTF-8 line
            yield b"data: {\"also_valid\": 2}\n\n"

        mock_response.aiter_bytes = mixed_chunks
        type(mock_response)._request = PropertyMock(return_value=None)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=dict,
            response=mock_response,
            client=mock_client,
        )

        events = []
        async for event in stream._iter_events():
            events.append(event)

        # Should successfully parse both valid events
        assert len(events) == 2
        assert events[0].json() == {"valid": 1}
        assert events[1].json() == {"also_valid": 2}


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



"""
Tests for production-hardened streaming fixes in AsyncStream.

These tests verify all critical production fixes:
1. Decoder timeout protection (prevents hanging on incomplete SSE events)
2. Max buffer size protection (prevents memory exhaustion)
3. Double-iteration guard (prevents resource issues)
4. Cleanup exception suppression (preserves original errors)
"""
import asyncio
from typing import Iterator, AsyncIterator, AsyncGenerator
from unittest.mock import Mock, PropertyMock

import httpx
import pytest

from sgp_agents._streaming import MAX_SSE_BUFFER_SIZE, Stream, SSEDecoder, AsyncStream


def create_mock_client() -> Mock:
    """Create a mock client with a real SSEDecoder."""
    mock_client = Mock()
    mock_client._make_sse_decoder.return_value = SSEDecoder()
    mock_client._process_response_data.return_value = {"processed": True}
    return mock_client


class TestDecoderTimeoutProtection:
    """Test that decoder-level timeout protection works correctly."""

    @pytest.mark.asyncio
    async def test_decoder_timeout_on_incomplete_sse_event(self) -> None:
        """
        CRITICAL: Verify timeout triggers when decoder waits for incomplete SSE event.

        This is the key production fix - previously, only byte iteration had timeout,
        but decoder could hang forever waiting for a terminating \\n\\n.
        """
        mock_response = Mock()

        async def incomplete_event_then_hang() -> AsyncIterator[bytes]:
            # Send incomplete SSE event (missing \n\n terminator)
            yield b'data: {"incomplete":'
            yield b' "value"}'
            # Server crashes or stops - no \n\n ever comes
            await asyncio.sleep(1000)

        mock_response.aiter_bytes = incomplete_event_then_hang

        # Configure timeout
        mock_request = Mock()
        timeout_dict = {'connect': 5.0, 'read': 0.5, 'write': 5.0, 'pool': 5.0}
        mock_request.extensions = {"timeout": timeout_dict}
        type(mock_response)._request = PropertyMock(return_value=mock_request)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=object,
            response=mock_response,
            client=mock_client,
        )

        # Should timeout in decoder, not hang forever
        with pytest.raises(httpx.ReadTimeout) as exc_info:
            async for _ in stream._iter_events():
                pass

        error_msg = str(exc_info.value)
        assert "SSE decoder timeout" in error_msg or "Stream read timeout" in error_msg

    @pytest.mark.asyncio
    async def test_decoder_completes_normally_with_proper_events(self) -> None:
        """Verify decoder doesn't false-positive timeout on normal events."""
        mock_response = Mock()

        async def normal_events() -> AsyncIterator[bytes]:
            yield b'data: {"event": 1}\n\n'
            await asyncio.sleep(0.1)
            yield b'data: {"event": 2}\n\n'

        mock_response.aiter_bytes = normal_events

        mock_request = Mock()
        timeout_dict = {'connect': 5.0, 'read': 2.0, 'write': 5.0, 'pool': 5.0}
        mock_request.extensions = {"timeout": timeout_dict}
        type(mock_response)._request = PropertyMock(return_value=mock_request)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=object,
            response=mock_response,
            client=mock_client,
        )

        events: list[object] = []
        async for event in stream._iter_events():
            events.append(event)

        assert len(events) == 2

    @pytest.mark.asyncio
    async def test_decoder_timeout_distinguished_from_byte_timeout(self) -> None:
        """Verify we can distinguish decoder timeout from byte-level timeout."""
        mock_response = Mock()

        async def bytes_arrive_but_incomplete() -> AsyncIterator[bytes]:
            # Bytes arrive promptly but never form complete event
            yield b'data: incomplete'
            await asyncio.sleep(0.1)
            yield b' more data'
            await asyncio.sleep(0.1)
            # No \n\n, decoder waits forever
            await asyncio.sleep(1000)

        mock_response.aiter_bytes = bytes_arrive_but_incomplete

        mock_request = Mock()
        timeout_dict = {'connect': 5.0, 'read': 0.3, 'write': 5.0, 'pool': 5.0}
        mock_request.extensions = {"timeout": timeout_dict}
        type(mock_response)._request = PropertyMock(return_value=mock_request)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=object,
            response=mock_response,
            client=mock_client,
        )

        with pytest.raises(httpx.ReadTimeout):
            async for _ in stream._iter_events():
                pass


class TestMaxBufferSizeProtection:
    """Test that SSE decoder enforces max buffer size."""

    @pytest.mark.asyncio
    async def test_decoder_rejects_oversized_buffer(self) -> None:
        """
        CRITICAL: Verify decoder raises error when buffer exceeds limit.

        This prevents memory exhaustion from malformed or malicious streams.
        """
        mock_response = Mock()

        async def huge_event_without_terminator() -> AsyncIterator[bytes]:
            # Send data that never terminates but grows indefinitely
            chunk_size = 1024 * 1024  # 1MB chunks
            for _ in range(110):  # 110MB total, exceeds 100MB limit
                yield b"x" * chunk_size

        mock_response.aiter_bytes = huge_event_without_terminator
        type(mock_response)._request = PropertyMock(return_value=None)

        # Create decoder with default max size
        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=object,
            response=mock_response,
            client=mock_client,
        )

        with pytest.raises(ValueError) as exc_info:
            async for _ in stream._iter_events():
                pass

        error_msg = str(exc_info.value)
        assert "SSE buffer exceeded maximum size" in error_msg
        assert str(MAX_SSE_BUFFER_SIZE) in error_msg

    @pytest.mark.asyncio
    async def test_decoder_accepts_large_valid_event(self) -> None:
        """Verify decoder accepts large events that are properly terminated."""
        mock_response = Mock()

        async def large_valid_event() -> AsyncIterator[bytes]:
            # Send 50MB event (under 100MB limit) with proper terminator
            large_data = "x" * (50 * 1024 * 1024)
            yield f'data: {{"large": "{large_data}"}}\n\n'.encode()

        mock_response.aiter_bytes = large_valid_event
        type(mock_response)._request = PropertyMock(return_value=None)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=object,
            response=mock_response,
            client=mock_client,
        )

        events: list[object] = []
        async for event in stream._iter_events():
            events.append(event)

        assert len(events) == 1

    @pytest.mark.asyncio
    async def test_decoder_custom_max_buffer_size(self) -> None:
        """Verify decoder respects custom max buffer size."""
        mock_response = Mock()

        async def medium_event() -> AsyncIterator[bytes]:
            # Send 500KB without terminator
            yield b"x" * (500 * 1024)

        mock_response.aiter_bytes = medium_event
        type(mock_response)._request = PropertyMock(return_value=None)

        # Create decoder with small custom max size
        mock_client = Mock()
        mock_client._make_sse_decoder.return_value = SSEDecoder(max_buffer_size=100 * 1024)  # 100KB
        mock_client._process_response_data.return_value = {"processed": True}

        stream = AsyncStream(
            cast_to=object,
            response=mock_response,
            client=mock_client,
        )

        with pytest.raises(ValueError) as exc_info:
            async for _ in stream._iter_events():
                pass

        assert "102400 bytes" in str(exc_info.value)  # 100KB in bytes


class TestDoubleIterationGuard:
    """Test that streams prevent double iteration."""

    @pytest.mark.asyncio
    async def test_async_stream_raises_on_double_iteration(self) -> None:
        """
        CRITICAL: Verify AsyncStream raises error on second iteration attempt.

        This prevents confusing connection errors and resource leaks.
        """
        mock_response = Mock()

        # Make aclose an async mock
        async def mock_aclose() -> None:
            pass
        mock_response.aclose = mock_aclose

        async def simple_event() -> AsyncIterator[bytes]:
            yield b'data: {"test": 1}\n\n'

        mock_response.aiter_bytes = simple_event
        type(mock_response)._request = PropertyMock(return_value=None)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=object,
            response=mock_response,
            client=mock_client,
        )

        # First iteration should work - use __aiter__ (which is what 'async for' calls)
        events1: list[object] = []
        async for event in stream:
            events1.append(event)
        assert len(events1) == 1

        # Second iteration should raise RuntimeError when trying to iterate again
        with pytest.raises(RuntimeError) as exc_info:
            async for _ in stream:
                pass

        assert "already been consumed" in str(exc_info.value)

    def test_sync_stream_raises_on_double_iteration(self) -> None:
        """Verify Stream (sync) also prevents double iteration."""
        # Create a simple mock response
        mock_response = Mock()
        mock_response.close = Mock()

        # Create a generator function that can be called multiple times
        def make_iter() -> Iterator[bytes]:
            yield b'data: {"test": 1}\n\n'

        mock_response.iter_bytes = make_iter

        mock_client = Mock()
        mock_client._make_sse_decoder.return_value = SSEDecoder()
        mock_client._process_response_data.return_value = {"processed": True}

        stream = Stream(
            cast_to=object,
            response=mock_response,
            client=mock_client,
        )

        # First iteration - use __iter__ which checks _consumed flag
        events1: list[object] = []
        for event in stream:
            events1.append(event)
        assert len(events1) == 1

        # Second iteration should raise
        with pytest.raises(RuntimeError) as exc_info:
            for _ in stream:
                pass

        assert "already been consumed" in str(exc_info.value)

    @pytest.mark.asyncio
    async def test_double_iteration_error_message_is_clear(self) -> None:
        """Verify error message provides clear guidance."""
        mock_response = Mock()

        async def mock_aclose() -> None:
            pass
        mock_response.aclose = mock_aclose

        async def simple_event() -> AsyncIterator[bytes]:
            yield b'data: {"test": 1}\n\n'

        mock_response.aiter_bytes = simple_event
        type(mock_response)._request = PropertyMock(return_value=None)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=object,
            response=mock_response,
            client=mock_client,
        )

        # Consume stream using __aiter__ which checks the flag
        async for _ in stream:
            pass

        # Try again
        with pytest.raises(RuntimeError) as exc_info:
            async for _ in stream:
                pass

        error_msg = str(exc_info.value)
        assert "AsyncStream" in error_msg
        assert "Create a new request" in error_msg


class TestCleanupExceptionSuppression:
    """Test that cleanup exceptions don't mask original errors."""

    @pytest.mark.asyncio
    async def test_cleanup_error_suppressed_preserves_timeout(self) -> None:
        """
        CRITICAL: Verify cleanup errors don't mask timeout errors.

        This is essential for production debugging - you need to see the
        real timeout, not a connection error during cleanup.
        """
        mock_response = Mock()
        cleanup_order: list[str] = []

        class FailingCleanupIterator:
            async def __anext__(self) -> bytes:
                await asyncio.sleep(1000)  # Will timeout
                return b"never"

            async def aclose(self) -> None:
                cleanup_order.append("aclose_called")
                # Simulate cleanup failing (e.g., connection already closed)
                raise ConnectionError("Connection already closed")

        mock_response.aiter_bytes = lambda: FailingCleanupIterator()

        mock_request = Mock()
        timeout_dict = {'connect': 5.0, 'read': 0.3, 'write': 5.0, 'pool': 5.0}
        mock_request.extensions = {"timeout": timeout_dict}
        type(mock_response)._request = PropertyMock(return_value=mock_request)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=object,
            response=mock_response,
            client=mock_client,
        )

        # Should raise ReadTimeout, NOT ConnectionError
        with pytest.raises(httpx.ReadTimeout) as exc_info:
            async for _ in stream._iter_events():
                pass

        # Verify it's the timeout error, not cleanup error
        assert "timeout" in str(exc_info.value).lower()
        assert "Connection already closed" not in str(exc_info.value)

        # Verify cleanup was attempted
        assert "aclose_called" in cleanup_order

    @pytest.mark.asyncio
    async def test_normal_cleanup_still_works(self) -> None:
        """Verify normal cleanup still executes properly."""
        mock_response = Mock()
        cleanup_tracker = {"aclose_called": False}

        async def normal_bytes() -> AsyncGenerator[bytes, None]:
            """Simple async generator that yields one event."""
            yield b'data: test\n\n'

        class TrackedIterator:
            """Iterator wrapper that tracks cleanup calls."""
            def __init__(self) -> None:
                self.gen: AsyncGenerator[bytes, None] = normal_bytes()

            async def __anext__(self) -> bytes:
                return await self.gen.__anext__()

            async def aclose(self) -> None:
                cleanup_tracker["aclose_called"] = True
                await self.gen.aclose()

        mock_response.aiter_bytes = TrackedIterator
        type(mock_response)._request = PropertyMock(return_value=None)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=object,
            response=mock_response,
            client=mock_client,
        )

        async for _ in stream._iter_events():
            pass

        # Verify cleanup was called
        assert cleanup_tracker["aclose_called"], "Cleanup should be called on normal completion"


class TestLoggingOptimization:
    """Test that logging has been optimized for production."""

    @pytest.mark.asyncio
    async def test_chunk_logging_reduced_frequency(self) -> None:
        """
        Verify chunk logging happens every 5000 chunks, not 1000.

        This reduces log volume by 80% for long-running streams.
        """
        mock_response = Mock()

        async def many_chunks() -> AsyncIterator[bytes]:
            for _ in range(6000):
                yield b'data: x\n\n'

        mock_response.aiter_bytes = many_chunks
        type(mock_response)._request = PropertyMock(return_value=None)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=object,
            response=mock_response,
            client=mock_client,
        )

        # This test mainly verifies code doesn't crash with many chunks
        # In production, you'd check logs to verify frequency
        event_count = 0
        async for _ in stream._iter_events():
            event_count += 1
            if event_count >= 100:  # Don't actually process all 6000
                break

        assert event_count == 100


class TestTimingImprovements:
    """Test that timing measurements are improved."""

    @pytest.mark.asyncio
    async def test_timing_consistency(self) -> None:
        """Verify timing measurements are consistent (no race conditions)."""
        mock_response = Mock()

        async def timed_chunks() -> AsyncIterator[bytes]:
            for i in range(3):
                await asyncio.sleep(0.1)
                yield f'data: {{"chunk": {i}}}\n\n'.encode()

        mock_response.aiter_bytes = timed_chunks

        mock_request = Mock()
        timeout_dict = {'connect': 5.0, 'read': 2.0, 'write': 5.0, 'pool': 5.0}
        mock_request.extensions = {"timeout": timeout_dict}
        type(mock_response)._request = PropertyMock(return_value=mock_request)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=object,
            response=mock_response,
            client=mock_client,
        )

        events: list[object] = []
        async for event in stream._iter_events():
            events.append(event)

        # Should complete without timing-related errors
        assert len(events) == 3


class TestBackwardCompatibility:
    """Test that changes maintain backward compatibility."""

    @pytest.mark.asyncio
    async def test_no_timeout_config_still_works(self) -> None:
        """Verify streams without timeout config still work (backward compat)."""
        mock_response = Mock()

        async def normal_chunks() -> AsyncIterator[bytes]:
            for i in range(3):
                await asyncio.sleep(0.1)
                yield f'data: {{"i": {i}}}\n\n'.encode()

        mock_response.aiter_bytes = normal_chunks
        type(mock_response)._request = PropertyMock(return_value=None)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=object,
            response=mock_response,
            client=mock_client,
        )

        events: list[object] = []
        async for event in stream._iter_events():
            events.append(event)

        assert len(events) == 3

    @pytest.mark.asyncio
    async def test_old_style_usage_patterns_work(self) -> None:
        """Verify old usage patterns still work."""
        mock_response = Mock()
        mock_response.aclose = Mock()

        async def simple() -> AsyncIterator[bytes]:
            yield b'data: {"old": "style"}\n\n'

        mock_response.aiter_bytes = simple
        type(mock_response)._request = PropertyMock(return_value=None)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=object,
            response=mock_response,
            client=mock_client,
        )

        # Old style: manually iterate _iter_events
        events: list[object] = []
        async for event in stream._iter_events():
            events.append(event)

        assert len(events) == 1


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



# Note: initially copied from https://github.com/florimondmanca/httpx-sse/blob/master/src/httpx_sse/_decoders.py
from __future__ import annotations

import json
import time
import asyncio
import inspect
import logging
from types import TracebackType
from typing import TYPE_CHECKING, Any, Generic, TypeVar, Iterator, AsyncIterator, cast
from typing_extensions import Self, Protocol, TypeGuard, override, get_origin, runtime_checkable

import httpx

from ._utils import extract_type_var_from_base

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from ._client import Agents, AsyncAgents


_T = TypeVar("_T")

# Maximum buffer size for SSE decoder to prevent memory exhaustion (100MB default)
# Production workloads have shown legitimate SSE events up to 44MB
MAX_SSE_BUFFER_SIZE = 100 * 1024 * 1024


class Stream(Generic[_T]):
    """Provides the core interface to iterate over a synchronous stream response."""

    response: httpx.Response

    _decoder: SSEBytesDecoder

    def __init__(
        self,
        *,
        cast_to: type[_T],
        response: httpx.Response,
        client: Agents,
    ) -> None:
        self.response = response
        self._cast_to = cast_to
        self._client = client
        self._decoder = client._make_sse_decoder()
        self._iterator = self.__stream__()
        self._consumed = False  # Guard against double iteration

    def __next__(self) -> _T:
        return self._iterator.__next__()

    def __iter__(self) -> Iterator[_T]:
        if self._consumed:
            raise RuntimeError(
                "Stream has already been consumed. "
                "Create a new request to iterate again."
            )
        self._consumed = True
        for item in self._iterator:
            yield item

    def _iter_events(self) -> Iterator[ServerSentEvent]:
        yield from self._decoder.iter_bytes(self.response.iter_bytes())

    def __stream__(self) -> Iterator[_T]:
        cast_to = cast(Any, self._cast_to)
        response = self.response
        process_data = self._client._process_response_data
        iterator = self._iter_events()

        try:
            for sse in iterator:
                yield process_data(data=sse.json(), cast_to=cast_to, response=response)
        finally:
            # Always close the response to release the connection, even if the
            # client breaks early or an exception occurs during streaming
            response.close()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.close()

    def close(self) -> None:
        """
        Close the response and release the connection.

        Automatically called if the response body is read to completion.
        """
        self.response.close()


class AsyncStream(Generic[_T]):
    """Provides the core interface to iterate over an asynchronous stream response."""

    response: httpx.Response

    _decoder: SSEDecoder | SSEBytesDecoder

    def __init__(
        self,
        *,
        cast_to: type[_T],
        response: httpx.Response,
        client: AsyncAgents,
    ) -> None:
        self.response = response
        self._cast_to = cast_to
        self._client = client
        self._decoder = client._make_sse_decoder()
        self._iterator = self.__stream__()
        self._consumed = False  # Guard against double iteration

    async def __anext__(self) -> _T:
        return await self._iterator.__anext__()

    async def __aiter__(self) -> AsyncIterator[_T]:
        if self._consumed:
            raise RuntimeError(
                "AsyncStream has already been consumed. "
                "Create a new request to iterate again."
            )
        self._consumed = True
        async for item in self._iterator:
            yield item

    async def _iter_events(self) -> AsyncIterator[ServerSentEvent]:
        """
        Iterate over SSE events with timeout protection at both the byte and decoder levels.
        This prevents indefinite hangs from incomplete SSE events or stalled connections.
        """
        # Extract timeout from request extensions
        # httpx stores timeout as a dict: {'connect': 5.0, 'read': 2.0, 'write': 5.0, 'pool': 5.0}
        # Default to None (no timeout) unless explicitly configured
        timeout_seconds: float | None = None
        timeout_extraction_debug: dict[str, Any] = {
            "has_request": False,
            "has_extensions": False,
            "timeout_dict": None
        }

        if hasattr(self.response, '_request') and self.response._request is not None:
            timeout_extraction_debug["has_request"] = True
            if hasattr(self.response._request, 'extensions'):
                timeout_extraction_debug["has_extensions"] = True
                # extensions is MutableMapping[str, Any] from httpx, safe to access as dict
                timeout_dict: Any = self.response._request.extensions.get('timeout')  # type: ignore[reportUnknownMemberType]
                timeout_extraction_debug["timeout_dict"] = timeout_dict

                # timeout_dict is a dictionary with timeout values
                if timeout_dict is not None and isinstance(timeout_dict, dict):
                    # Extract read timeout from dict
                    timeout_dict_typed = cast("dict[str, Any]", timeout_dict)
                    read_timeout = timeout_dict_typed.get('read')
                    if read_timeout is not None and isinstance(read_timeout, (int, float)):
                        timeout_seconds = float(read_timeout)
                        logger.debug(
                            f"[AsyncStream] Timeout extraction successful: read_timeout={timeout_seconds}s"
                        )
                    else:
                        logger.debug(
                            f"[AsyncStream] Read timeout is None in config. "
                            "No per-chunk timeout will be applied."
                        )
                else:
                    logger.warning(
                        f"[AsyncStream] Timeout config exists but is not a dict: "
                        f"type={type(timeout_dict)}, value={timeout_dict}"
                    )

        if timeout_seconds is None:
            logger.debug(
                f"[AsyncStream] No timeout configured. "
                "Stream will wait indefinitely for chunks (original behavior)."
            )

        # Wrap byte iterator with timeout protection while maintaining single stream
        chunk_count = 0
        total_bytes = 0
        start_time = time.time()
        last_chunk_time = start_time

        async def timeout_protected_bytes() -> AsyncIterator[bytes]:
            """
            Wraps byte iteration with per-chunk timeout while keeping the stream continuous.
            Cleanup exceptions are suppressed to preserve the original error.
            """
            nonlocal chunk_count, total_bytes, last_chunk_time

            byte_iterator = self.response.aiter_bytes()

            try:
                while True:
                    try:
                        chunk_start = time.time()

                        if timeout_seconds is not None:
                            # Apply timeout to each chunk read
                            chunk = await asyncio.wait_for(
                                byte_iterator.__anext__(),
                                timeout=timeout_seconds
                            )
                        else:
                            # No timeout - use original behavior
                            chunk = await byte_iterator.__anext__()

                        chunk_count += 1
                        chunk_size = len(chunk)
                        total_bytes += chunk_size
                        chunk_duration = time.time() - chunk_start

                        # Update timing after all measurements
                        time_since_last = time.time() - last_chunk_time
                        last_chunk_time = time.time()

                        # Log periodically (every 5000 chunks) and for slow chunks
                        if chunk_count % 5000 == 0:
                            logger.debug(
                                f"[AsyncStream] Chunk #{chunk_count}: "
                                f"size={chunk_size}B, "
                                f"duration={chunk_duration:.2f}s, "
                                f"since_last={time_since_last:.2f}s, "
                                f"total_bytes={total_bytes}"
                            )
                        elif chunk_duration > 10.0:
                            logger.debug(
                                f"[AsyncStream] Slow chunk #{chunk_count}: "
                                f"duration={chunk_duration:.2f}s"
                            )

                        yield chunk

                    except StopAsyncIteration:
                        elapsed = time.time() - start_time
                        logger.info(
                            f"[AsyncStream] Stream completed normally: "
                            f"chunks={chunk_count}, "
                            f"total_bytes={total_bytes}, "
                            f"elapsed={elapsed:.2f}s"
                        )
                        break

                    except asyncio.TimeoutError as timeout_err:
                        elapsed_since_last = time.time() - last_chunk_time
                        total_elapsed = time.time() - start_time

                        logger.error(
                            f"[AsyncStream] TIMEOUT after {elapsed_since_last:.1f}s waiting for byte chunk: "
                            f"last_chunk=#{chunk_count}, "
                            f"total_bytes={total_bytes}, "
                            f"timeout_config={timeout_seconds}s, "
                            f"total_elapsed={total_elapsed:.1f}s. "
                            f"Server may have stopped sending data without closing connection."
                        )

                        # Raise httpx.ReadTimeout so application code can catch it specifically
                        raise httpx.ReadTimeout(
                            f"Stream read timeout after {timeout_seconds}s waiting for data. "
                            f"Last chunk received: #{chunk_count} ({total_bytes} bytes total). "
                            "Server may have stopped sending data without closing connection."
                        ) from timeout_err

            except Exception as e:
                # RemoteProtocolError is expected when server closes connection abruptly
                # Check by class name as well to handle import variations
                is_remote_protocol_error = (
                    isinstance(e, httpx.RemoteProtocolError) or
                    type(e).__name__ == 'RemoteProtocolError'
                )
                is_expected_exception = (
                    isinstance(e, (StopAsyncIteration, httpx.ReadTimeout, asyncio.TimeoutError)) or
                    type(e).__name__ in ('StopAsyncIteration', 'ReadTimeout', 'TimeoutError')
                )

                if is_remote_protocol_error:
                    logger.info(
                        f"[AsyncStream] Server closed connection: "
                        f"chunk_count={chunk_count}, "
                        f"total_bytes={total_bytes}"
                    )
                elif not is_expected_exception:
                    logger.error(
                        f"[AsyncStream] Unexpected error in byte iterator: "
                        f"type={type(e).__name__}, "
                        f"error={e}, "
                        f"chunk_count={chunk_count}, "
                        f"total_bytes={total_bytes}"
                    )
                raise
            finally:
                # Suppress cleanup exceptions to preserve the original error
                try:
                    if hasattr(byte_iterator, 'aclose') and callable(getattr(byte_iterator, 'aclose', None)):
                        await byte_iterator.aclose()  # type: ignore[attr-defined]
                        logger.debug("[AsyncStream] Byte iterator cleanup completed")
                except Exception as cleanup_error:
                    # Suppress cleanup errors to preserve the original exception
                    logger.warning(
                        f"[AsyncStream] Error during byte iterator cleanup (suppressed): "
                        f"{type(cleanup_error).__name__}: {cleanup_error}"
                    )

        # Apply timeout to both byte iterator and decoder to prevent hangs when
        # waiting for complete SSE events (e.g., missing \n\n terminator)
        sse_count = 0
        try:
            # Create decoder iterator
            decoder_iterator = self._decoder.aiter_bytes(timeout_protected_bytes())

            # Apply timeout to decoder iteration as well
            while True:
                try:
                    if timeout_seconds is not None:
                        sse = await asyncio.wait_for(
                            decoder_iterator.__anext__(),
                            timeout=timeout_seconds
                        )
                    else:
                        sse = await decoder_iterator.__anext__()

                    sse_count += 1

                    if sse_count % 5000 == 0:
                        logger.debug(
                            f"[AsyncStream] Decoded SSE event #{sse_count}: "
                            f"total_chunks={chunk_count}, "
                            f"total_bytes={total_bytes}"
                        )

                    yield sse

                except StopAsyncIteration:
                    logger.info(
                        f"[AsyncStream] All SSE events decoded: "
                        f"count={sse_count}, "
                        f"chunks={chunk_count}, "
                        f"bytes={total_bytes}"
                    )
                    break

                except asyncio.TimeoutError as timeout_err:
                    logger.error(
                        f"[AsyncStream] TIMEOUT in SSE decoder after {timeout_seconds}s. "
                        f"Last SSE event: #{sse_count}, "
                        f"chunks={chunk_count}, "
                        f"bytes={total_bytes}. "
                        "Decoder may be waiting for incomplete SSE event (missing \\n\\n terminator)."
                    )
                    raise httpx.ReadTimeout(
                        f"SSE decoder timeout after {timeout_seconds}s. "
                        f"Processed {sse_count} events, {chunk_count} chunks, {total_bytes} bytes. "
                        "Server may have sent incomplete SSE event."
                    ) from timeout_err

        except Exception as e:
            # RemoteProtocolError and ValueError (buffer overflow) are expected, log at INFO
            # Check by class name as well to handle import variations
            is_expected_error = (
                isinstance(e, (httpx.RemoteProtocolError, ValueError)) or
                type(e).__name__ in ('RemoteProtocolError', 'ValueError')
            )

            if is_expected_error:
                logger.info(
                    f"[AsyncStream] Stream ended: "
                    f"type={type(e).__name__}, "
                    f"sse_count={sse_count}, "
                    f"chunk_count={chunk_count}, "
                    f"total_bytes={total_bytes}"
                )
            else:
                logger.error(
                    f"[AsyncStream] Error during SSE decoding: "
                    f"type={type(e).__name__}, "
                    f"error={e}, "
                    f"sse_count={sse_count}, "
                    f"chunk_count={chunk_count}, "
                    f"total_bytes={total_bytes}"
                )
            raise

    async def __stream__(self) -> AsyncIterator[_T]:
        cast_to = cast(Any, self._cast_to)
        response = self.response
        process_data = self._client._process_response_data
        iterator = self._iter_events()

        try:
            async for sse in iterator:
                yield process_data(data=sse.json(), cast_to=cast_to, response=response)
        finally:
            # Always close the response to release the connection, even if the
            # client breaks early or an exception occurs during streaming
            await response.aclose()

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.close()

    async def close(self) -> None:
        """
        Close the response and release the connection.

        Automatically called if the response body is read to completion.
        """
        await self.response.aclose()


class ServerSentEvent:
    def __init__(
        self,
        *,
        event: str | None = None,
        data: str | None = None,
        id: str | None = None,
        retry: int | None = None,
    ) -> None:
        if data is None:
            data = ""

        self._id = id
        self._data = data
        self._event = event or None
        self._retry = retry

    @property
    def event(self) -> str | None:
        return self._event

    @property
    def id(self) -> str | None:
        return self._id

    @property
    def retry(self) -> int | None:
        return self._retry

    @property
    def data(self) -> str:
        return self._data

    def json(self) -> Any:
        try:
            return json.loads(self.data)
        except json.JSONDecodeError as e:
            logger.error(
                f"[SSE] Invalid JSON in event data: {e}. "
                f"Data preview: {self.data[:200] if len(self.data) > 200 else self.data}"
            )
            raise ValueError(f"SSE event contains invalid JSON: {e}") from e

    @override
    def __repr__(self) -> str:
        return f"ServerSentEvent(event={self.event}, data={self.data}, id={self.id}, retry={self.retry})"


class SSEDecoder:
    """
    SSE decoder with max buffer size protection to prevent memory exhaustion
    from malformed streams that never send proper event terminators.
    """
    _data: list[str]
    _event: str | None
    _retry: int | None
    _last_event_id: str | None
    _max_buffer_size: int

    def __init__(self, max_buffer_size: int = MAX_SSE_BUFFER_SIZE) -> None:
        self._event = None
        self._data = []
        self._last_event_id = None
        self._retry = None
        self._max_buffer_size = max_buffer_size

    def iter_bytes(self, iterator: Iterator[bytes]) -> Iterator[ServerSentEvent]:
        """Given an iterator that yields raw binary data, iterate over it & yield every event encountered"""
        for chunk in self._iter_chunks(iterator):
            # Split before decoding so splitlines() only uses \r and \n
            for raw_line in chunk.splitlines():
                try:
                    line = raw_line.decode("utf-8")
                except UnicodeDecodeError as e:
                    logger.warning(
                        f"[SSE] Failed to decode line as UTF-8: {e}. "
                        f"Line preview (hex): {raw_line[:50].hex()}"
                    )
                    # Skip malformed lines rather than crash the entire stream
                    continue
                sse = self.decode(line)
                if sse:
                    yield sse

    def _iter_chunks(self, iterator: Iterator[bytes]) -> Iterator[bytes]:
        """Given an iterator that yields raw binary data, iterate over it and yield individual SSE chunks."""
        data = b""
        for chunk in iterator:
            for line in chunk.splitlines(keepends=True):
                data += line

                # Check buffer size to prevent memory exhaustion
                if len(data) > self._max_buffer_size:
                    raise ValueError(
                        f"SSE buffer exceeded maximum size of {self._max_buffer_size} bytes "
                        "without receiving proper event terminator (\\n\\n). "
                        "This may indicate a malformed stream or missing event delimiters."
                    )

                if data.endswith((b"\r\r", b"\n\n", b"\r\n\r\n")):
                    yield data
                    data = b""
        if data:
            yield data

    async def aiter_bytes(self, iterator: AsyncIterator[bytes]) -> AsyncIterator[ServerSentEvent]:
        """Given an iterator that yields raw binary data, iterate over it & yield every event encountered"""
        async for chunk in self._aiter_chunks(iterator):
            # Split before decoding so splitlines() only uses \r and \n
            for raw_line in chunk.splitlines():
                try:
                    line = raw_line.decode("utf-8")
                except UnicodeDecodeError as e:
                    logger.warning(
                        f"[SSE] Failed to decode line as UTF-8: {e}. "
                        f"Line preview (hex): {raw_line[:50].hex()}"
                    )
                    # Skip malformed lines rather than crash the entire stream
                    continue
                sse = self.decode(line)
                if sse:
                    yield sse

    async def _aiter_chunks(self, iterator: AsyncIterator[bytes]) -> AsyncIterator[bytes]:
        """Given an async iterator that yields raw binary data, iterate over it and yield individual SSE chunks."""
        data = b""
        async for chunk in iterator:
            for line in chunk.splitlines(keepends=True):
                data += line

                # Check buffer size to prevent memory exhaustion
                if len(data) > self._max_buffer_size:
                    raise ValueError(
                        f"SSE buffer exceeded maximum size of {self._max_buffer_size} bytes "
                        "without receiving proper event terminator (\\n\\n). "
                        "This may indicate a malformed stream or missing event delimiters."
                    )

                if data.endswith((b"\r\r", b"\n\n", b"\r\n\r\n")):
                    yield data
                    data = b""
        if data:
            yield data

    def decode(self, line: str) -> ServerSentEvent | None:
        # See: https://html.spec.whatwg.org/multipage/server-sent-events.html#event-stream-interpretation  # noqa: E501

        if not line:
            if not self._event and not self._data and not self._last_event_id and self._retry is None:
                return None

            sse = ServerSentEvent(
                event=self._event,
                data="\n".join(self._data),
                id=self._last_event_id,
                retry=self._retry,
            )

            # NOTE: as per the SSE spec, do not reset last_event_id.
            self._event = None
            self._data = []
            self._retry = None

            return sse

        if line.startswith(":"):
            return None

        fieldname, _, value = line.partition(":")

        if value.startswith(" "):
            value = value[1:]

        if fieldname == "event":
            self._event = value
        elif fieldname == "data":
            self._data.append(value)
        elif fieldname == "id":
            if "\0" in value:
                pass
            else:
                self._last_event_id = value
        elif fieldname == "retry":
            try:
                self._retry = int(value)
            except (TypeError, ValueError):
                pass
        else:
            pass  # Field is ignored.

        return None


@runtime_checkable
class SSEBytesDecoder(Protocol):
    def iter_bytes(self, iterator: Iterator[bytes]) -> Iterator[ServerSentEvent]:
        """Given an iterator that yields raw binary data, iterate over it & yield every event encountered"""
        ...

    def aiter_bytes(self, iterator: AsyncIterator[bytes]) -> AsyncIterator[ServerSentEvent]:
        """Given an async iterator that yields raw binary data, iterate over it & yield every event encountered"""
        ...


def is_stream_class_type(typ: type) -> TypeGuard[type[Stream[object]] | type[AsyncStream[object]]]:
    """TypeGuard for determining whether or not the given type is a subclass of `Stream` / `AsyncStream`"""
    origin = get_origin(typ) or typ
    return inspect.isclass(origin) and issubclass(origin, (Stream, AsyncStream))


def extract_stream_chunk_type(
    stream_cls: type,
    *,
    failure_message: str | None = None,
) -> type:
    """Given a type like `Stream[T]`, returns the generic type variable `T`.

    This also handles the case where a concrete subclass is given, e.g.
    ```py
    class MyStream(Stream[bytes]):
        ...

    extract_stream_chunk_type(MyStream) -> bytes
    ```
    """
    from ._base_client import Stream, AsyncStream

    return extract_type_var_from_base(
        stream_cls,
        index=0,
        generic_bases=cast("tuple[type, ...]", (Stream, AsyncStream)),
        failure_message=failure_message,
    )



"""
Tests for corrected streaming timeout fix in AsyncStream._iter_events()

These tests verify that AsyncStream properly times out when the server
stops sending data, while preserving decoder state for multi-chunk events.
"""
# pyright: reportUnknownVariableType=false, reportUnknownMemberType=false, reportUnknownArgumentType=false
import asyncio
from typing import Any, AsyncIterator
from unittest.mock import Mock, PropertyMock

import httpx
import pytest

from sgp_agents._constants import DEFAULT_TIMEOUT
from sgp_agents._streaming import SSEDecoder, AsyncStream


def create_mock_client() -> Mock:
    """Create a mock client with a real SSEDecoder."""
    mock_client = Mock()
    mock_client._make_sse_decoder.return_value = SSEDecoder()
    return mock_client


class TestStreamingTimeoutCorrected:
    """Test suite for corrected streaming timeout behavior."""

    @pytest.mark.asyncio
    async def test_timeout_extraction_from_httpx_timeout_dict(self) -> None:
        """Verify timeout is correctly extracted from httpx timeout dict."""
        mock_response = Mock()

        async def normal_chunks() -> AsyncIterator[bytes]:
            for i in range(3):
                await asyncio.sleep(0.01)
                yield f"data: chunk{i}\n\n".encode()

        mock_response.aiter_bytes = normal_chunks

        # Mock request with timeout dict (how httpx actually stores it)
        mock_request = Mock()
        timeout_dict = {'connect': 5.0, 'read': 2.0, 'write': 5.0, 'pool': 5.0}
        mock_request.extensions = {"timeout": timeout_dict}
        type(mock_response)._request = PropertyMock(return_value=mock_request)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=dict,
            response=mock_response,
            client=mock_client,
        )

        # Should extract read timeout correctly without crashing
        events = []
        async for event in stream._iter_events():
            events.append(event)

        assert len(events) == 3

    @pytest.mark.asyncio
    async def test_timeout_none_means_no_timeout(self) -> None:
        """Verify that timeout dict with read=None means no timeout applied."""
        mock_response = Mock()

        async def slow_chunks() -> AsyncIterator[bytes]:
            """Chunks arrive slowly but should complete without timeout."""
            for i in range(2):
                await asyncio.sleep(0.5)  # Would timeout if 0.3s timeout applied
                yield f"data: chunk{i}\n\n".encode()

        mock_response.aiter_bytes = slow_chunks

        # Timeout dict with read=None means no read timeout
        mock_request = Mock()
        timeout_dict = {'connect': 5.0, 'read': None, 'write': 5.0, 'pool': 5.0}
        mock_request.extensions = {"timeout": timeout_dict}
        type(mock_response)._request = PropertyMock(return_value=mock_request)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=dict,
            response=mock_response,
            client=mock_client,
        )

        # Should complete without timing out
        events = []
        async for event in stream._iter_events():
            events.append(event)

        assert len(events) == 2

    @pytest.mark.asyncio
    async def test_decoder_state_preserved_across_chunks(self) -> None:
        """Verify decoder buffer is maintained when events span multiple chunks."""
        mock_response = Mock()

        async def split_event_chunks() -> AsyncIterator[bytes]:
            """Simulate SSE event split across multiple chunks."""
            # First chunk: incomplete event (no \n\n terminator)
            yield b'data: {"key":'
            await asyncio.sleep(0.01)
            # Second chunk: completes the event
            yield b'"value"}\n\n'

        mock_response.aiter_bytes = split_event_chunks

        # No timeout configured
        type(mock_response)._request = PropertyMock(return_value=None)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=dict,
            response=mock_response,
            client=mock_client,
        )

        # Should receive ONE complete event (decoder buffered across chunks)
        events = []
        async for event in stream._iter_events():
            events.append(event)

        assert len(events) == 1
        assert events[0].data == '{"key":"value"}'

    @pytest.mark.asyncio
    async def test_multi_byte_utf8_across_chunks(self) -> None:
        """Verify multi-byte UTF-8 characters split across chunks work correctly."""
        mock_response = Mock()

        async def split_utf8_chunks() -> AsyncIterator[bytes]:
            """Split a multi-byte UTF-8 character across chunks."""
            # Unicode character '你' (U+4F60) in UTF-8: 0xE4 0xBD 0xA0
            # Split it across chunks
            yield b'data: hello '
            yield b'\xe4'  # First byte of 你
            yield b'\xbd\xa0'  # Remaining bytes of 你
            yield b'\n\n'

        mock_response.aiter_bytes = split_utf8_chunks
        type(mock_response)._request = PropertyMock(return_value=None)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=dict,
            response=mock_response,
            client=mock_client,
        )

        events = []
        async for event in stream._iter_events():
            events.append(event)

        assert len(events) == 1
        assert events[0].data == 'hello 你'

    @pytest.mark.asyncio
    async def test_timeout_raises_when_server_hangs(self) -> None:
        """Verify timeout occurs when server stops sending."""
        mock_response = Mock()

        async def hung_after_one() -> AsyncIterator[bytes]:
            yield b"data: chunk0\n\n"
            await asyncio.sleep(1000)  # Hang

        mock_response.aiter_bytes = hung_after_one

        # Short timeout dict
        mock_request = Mock()
        timeout_dict = {'connect': 5.0, 'read': 0.5, 'write': 5.0, 'pool': 5.0}
        mock_request.extensions = {"timeout": timeout_dict}
        type(mock_response)._request = PropertyMock(return_value=mock_request)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=dict,
            response=mock_response,
            client=mock_client,
        )

        events = []
        with pytest.raises(httpx.ReadTimeout) as exc_info:
            async for event in stream._iter_events():
                events.append(event)

        assert len(events) == 1
        # Updated: Now catches timeout in decoder layer, not just byte layer
        error_msg = str(exc_info.value)
        assert ("Stream read timeout after 0.5s" in error_msg or
                "SSE decoder timeout after 0.5s" in error_msg)

    @pytest.mark.asyncio
    async def test_cleanup_on_timeout(self) -> None:
        """Verify cleanup happens on timeout."""
        mock_response = Mock()
        cleanup_tracker = {'aclose_called': False}

        class TrackedIterator:
            async def __anext__(self) -> bytes:
                await asyncio.sleep(1000)
                return b"never"

            async def aclose(self) -> None:
                cleanup_tracker['aclose_called'] = True

        mock_response.aiter_bytes = lambda: TrackedIterator()

        mock_request = Mock()
        timeout_dict = {'connect': 5.0, 'read': 0.3, 'write': 5.0, 'pool': 5.0}
        mock_request.extensions = {"timeout": timeout_dict}
        type(mock_response)._request = PropertyMock(return_value=mock_request)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=dict,
            response=mock_response,
            client=mock_client,
        )

        with pytest.raises(httpx.ReadTimeout):
            async for _ in stream._iter_events():
                pass

        assert cleanup_tracker['aclose_called'], "Should call aclose on timeout"

    @pytest.mark.asyncio
    async def test_event_loop_responsive_during_wait(self) -> None:
        """Verify event loop can run other tasks while waiting."""
        mock_response = Mock()

        async def slow_chunks() -> AsyncIterator[bytes]:
            for i in range(3):
                await asyncio.sleep(0.3)
                yield f"data: chunk{i}\n\n".encode()

        mock_response.aiter_bytes = slow_chunks

        mock_request = Mock()
        timeout_dict = {'connect': 5.0, 'read': 2.0, 'write': 5.0, 'pool': 5.0}
        mock_request.extensions = {"timeout": timeout_dict}
        type(mock_response)._request = PropertyMock(return_value=mock_request)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=dict,
            response=mock_response,
            client=mock_client,
        )

        # Track watchdog runs
        watchdog_ticks = []

        async def watchdog() -> None:
            for i in range(10):
                await asyncio.sleep(0.1)
                watchdog_ticks.append(i)

        async def consume() -> list[Any]:  # type: ignore[reportUnknownParameterType]
            events: list[Any] = []
            async for event in stream._iter_events():
                events.append(event)
            return events

        # Run both concurrently
        stream_task = asyncio.create_task(consume())
        watchdog_task = asyncio.create_task(watchdog())

        events = await stream_task
        watchdog_task.cancel()

        assert len(events) == 3
        assert len(watchdog_ticks) > 0, "Watchdog should run (event loop responsive)"

    @pytest.mark.asyncio
    async def test_no_timeout_when_extensions_missing(self) -> None:
        """Verify no timeout applied when request has no extensions."""
        mock_response = Mock()

        async def chunks() -> AsyncIterator[bytes]:
            for i in range(2):
                await asyncio.sleep(0.5)
                yield f"data: chunk{i}\n\n".encode()

        mock_response.aiter_bytes = chunks

        # No request or extensions
        type(mock_response)._request = PropertyMock(return_value=None)

        mock_client = create_mock_client()
        stream = AsyncStream(
            cast_to=dict,
            response=mock_response,
            client=mock_client,
        )

        # Should complete without timeout (no timeout configured)
        events = []
        async for event in stream._iter_events():
            events.append(event)

        assert len(events) == 2

    def test_default_timeout_is_300_seconds(self) -> None:
        """Verify that the default timeout is set to 300 seconds (5 minutes)."""
        # This test ensures that the timeout change for agent operations is preserved
        assert DEFAULT_TIMEOUT.connect == 5.0, "Connect timeout should remain 5.0s"
        assert DEFAULT_TIMEOUT.read == 300.0, "Read timeout should be 300.0s (5 minutes)"
        assert DEFAULT_TIMEOUT.write == 300.0, "Write timeout should be 300.0s (5 minutes)"
        assert DEFAULT_TIMEOUT.pool == 300.0, "Pool timeout should be 300.0s (5 minutes)"


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Union, Optional
from typing_extensions import Required, TypedDict

__all__ = ["WorkflowItemParam"]


class WorkflowItemParam(TypedDict, total=False):
    workflow_name: Required[str]

    workflow_alias: Optional[str]

    workflow_inputs: Optional[Dict[str, Union[str, Dict[str, Union[str, object]]]]]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, List, Union, Iterable, Optional
from typing_extensions import Literal, Required, TypeAlias, TypedDict

from .jinja_node_template_param import JinjaNodeTemplateParam
from .compound_condition_input_param import CompoundConditionInputParam

__all__ = [
    "NodeItemInputParam",
    "Config",
    "ConfigNodeConfig",
    "ConfigBatchedNodeConfig",
    "ConfigChunkEvaluationNodeConfig",
    "ConfigRerankerNodeConfig",
    "ConfigRetrieverNodeConfig",
    "ConfigCitationNodeConfig",
    "ConfigCitationNodeConfigCitationContext",
    "ConfigSearchCitationNodeConfig",
    "ConfigCodeExecutionConfig",
    "ConfigConditionNodeConfigInput",
    "ConfigDataTransformNodeConfig",
    "ConfigCreateMessagesNodeConfig",
    "ConfigCreateMessagesNodeConfigMessageConfig",
    "ConfigCreateMessagesNodeConfigMessageConfigAlternatingRoleMessages",
    "ConfigCreateMessagesNodeConfigMessageConfigSingleRoleMessages",
    "ConfigInsertMessagesConfig",
    "ConfigRemoveMessageConfig",
    "ConfigGetMessageConfig",
    "ConfigTokenizerChatTemplateConfig",
    "ConfigGenerationNodeConfig",
    "ConfigGenerationNodeConfigRetryConfig",
    "ConfigChatGenerationNodeConfig",
    "ConfigChatGenerationNodeConfigRetryConfig",
    "ConfigGenerationWithCitationsNodeConfig",
    "ConfigGenerationWithCitationsNodeConfigRetryConfig",
    "ConfigChatGenerationWithCitationsNodeConfig",
    "ConfigChatGenerationWithCitationsNodeConfigRetryConfig",
    "ConfigResponseParserNodeConfig",
    "ConfigJinjaNodeConfig",
    "ConfigProcessingNodeConfig",
    "ConfigProcessingNodeConfigFunctionSpecs",
    "ConfigRegexMatchNodeConfig",
    "ConfigSqlExecutorNodeConfig",
    "ConfigStaticNodeConfig",
    "ConfigLlmEngineNodeConfig",
    "ConfigLlmEngineNodeConfigBatchSysKwargs",
]


class ConfigNodeConfig(TypedDict, total=False):
    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    type_hints: Optional[object]


class ConfigBatchedNodeConfig(TypedDict, total=False):
    batch_size: Required[int]

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    split_approx_evenly: bool

    type_hints: Optional[object]


class ConfigChunkEvaluationNodeConfig(TypedDict, total=False):
    top_k_thresholds: Required[Iterable[int]]

    fuzzy_match_threshold: float

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    require_all: bool

    type_hints: Optional[object]


class ConfigRerankerNodeConfig(TypedDict, total=False):
    num_to_return: Required[int]

    scorers: Required[Iterable[object]]

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    score_threshold: Optional[float]

    type_hints: Optional[object]


class ConfigRetrieverNodeConfig(TypedDict, total=False):
    num_to_return: Required[int]

    exact_knn_search: Optional[bool]

    knowledge_base_id: Optional[str]

    knowledge_base_name: Optional[str]

    metadata: Optional[Dict[str, Optional[str]]]

    min_results_per_knowledge_base: int

    node_metadata: Optional[List[str]]

    num_retriever_workers: int

    num_workers: Optional[int]

    type_hints: Optional[object]


class ConfigCitationNodeConfigCitationContext(TypedDict, total=False):
    generate_with_llm: bool

    metric: Optional[str]

    min_similarity: Optional[float]

    score: Optional[Literal["precision", "recall", "fmeasure"]]


class ConfigCitationNodeConfig(TypedDict, total=False):
    citation_type: Required[Literal["rouge", "model_defined"]]

    citation_context: ConfigCitationNodeConfigCitationContext

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    s3_path_override: Optional[str]

    type_hints: Optional[object]


class ConfigSearchCitationNodeConfig(TypedDict, total=False):
    end_search_regex: Required[str]

    search_regex: Required[str]

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    type_hints: Optional[object]


class ConfigCodeExecutionConfig(TypedDict, total=False):
    files: Required[Dict[str, str]]

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    output_artifacts_dir: Optional[str]

    type_hints: Optional[object]


class ConfigConditionNodeConfigInput(TypedDict, total=False):
    condition: Required[CompoundConditionInputParam]
    """Representation of a compound boolean statement, i.e.

    a negation, conjunction, or disjunction of UnaryConditions
    """

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    type_hints: Optional[object]


class ConfigDataTransformNodeConfig(TypedDict, total=False):
    action: Required[str]

    additional_inputs: object

    apply_to_dictlist_leaves: bool

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    type_hints: Optional[object]


class ConfigCreateMessagesNodeConfigMessageConfigAlternatingRoleMessages(TypedDict, total=False):
    role_value_pairs: Required[Iterable[Dict[str, str]]]


class ConfigCreateMessagesNodeConfigMessageConfigSingleRoleMessages(TypedDict, total=False):
    content: Required[str]

    role: Required[str]


ConfigCreateMessagesNodeConfigMessageConfig: TypeAlias = Union[
    ConfigCreateMessagesNodeConfigMessageConfigAlternatingRoleMessages,
    ConfigCreateMessagesNodeConfigMessageConfigSingleRoleMessages,
]


class ConfigCreateMessagesNodeConfig(TypedDict, total=False):
    message_configs: Required[Iterable[ConfigCreateMessagesNodeConfigMessageConfig]]

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    type_hints: Optional[object]


class ConfigInsertMessagesConfig(TypedDict, total=False):
    index: Required[int]

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    type_hints: Optional[object]


class ConfigRemoveMessageConfig(TypedDict, total=False):
    index: Required[int]

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    type_hints: Optional[object]


class ConfigGetMessageConfig(TypedDict, total=False):
    index: Required[int]

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    type_hints: Optional[object]


class ConfigTokenizerChatTemplateConfig(TypedDict, total=False):
    llm_model: Required[str]

    add_generation_prompt: bool

    kwargs: object

    max_length: Optional[int]

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    padding: bool

    truncation: bool

    type_hints: Optional[object]


class ConfigGenerationNodeConfigRetryConfig(TypedDict, total=False):
    backoff: int

    delay: int

    exceptions: List[Literal["SGPClientError", "APITimeoutError", "InternalServerError", "RateLimitError", "Exception"]]

    tries: int


class ConfigGenerationNodeConfig(TypedDict, total=False):
    llm_model: str

    llm_model_deployment: Optional[str]

    llm_model_instance: Optional[str]

    max_tokens: int

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    retry_config: ConfigGenerationNodeConfigRetryConfig

    stop_sequences: Optional[List[str]]

    strip_whitespace: bool

    temperature: float

    tool_name: Optional[str]

    type_hints: Optional[object]


class ConfigChatGenerationNodeConfigRetryConfig(TypedDict, total=False):
    backoff: int

    delay: int

    exceptions: List[Literal["SGPClientError", "APITimeoutError", "InternalServerError", "RateLimitError", "Exception"]]

    tries: int


class ConfigChatGenerationNodeConfig(TypedDict, total=False):
    memory_strategy: Required[object]

    instructions: Optional[str]

    llm_model: str

    llm_model_deployment: Optional[str]

    llm_model_instance: Optional[str]

    max_tokens: int

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    retry_config: ConfigChatGenerationNodeConfigRetryConfig

    stop_sequences: Optional[List[str]]

    strip_whitespace: bool

    temperature: float

    tool_name: Optional[str]

    type_hints: Optional[object]


class ConfigGenerationWithCitationsNodeConfigRetryConfig(TypedDict, total=False):
    backoff: int

    delay: int

    exceptions: List[Literal["SGPClientError", "APITimeoutError", "InternalServerError", "RateLimitError", "Exception"]]

    tries: int


class ConfigGenerationWithCitationsNodeConfig(TypedDict, total=False):
    regex_pattern: Required[str]

    regex_replace: Required[str]

    llm_model: str

    llm_model_deployment: Optional[str]

    llm_model_instance: Optional[str]

    max_tokens: int

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    retry_config: ConfigGenerationWithCitationsNodeConfigRetryConfig

    stop_sequences: Optional[List[str]]

    strip_whitespace: bool

    temperature: float

    tool_name: Optional[str]

    type_hints: Optional[object]


class ConfigChatGenerationWithCitationsNodeConfigRetryConfig(TypedDict, total=False):
    backoff: int

    delay: int

    exceptions: List[Literal["SGPClientError", "APITimeoutError", "InternalServerError", "RateLimitError", "Exception"]]

    tries: int


class ConfigChatGenerationWithCitationsNodeConfig(TypedDict, total=False):
    memory_strategy: Required[object]

    regex_pattern: Required[str]

    regex_replace: Required[str]

    instructions: Optional[str]

    llm_model: str

    llm_model_deployment: Optional[str]

    llm_model_instance: Optional[str]

    max_tokens: int

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    retry_config: ConfigChatGenerationWithCitationsNodeConfigRetryConfig

    stop_sequences: Optional[List[str]]

    strip_whitespace: bool

    temperature: float

    tool_name: Optional[str]

    type_hints: Optional[object]


class ConfigResponseParserNodeConfig(TypedDict, total=False):
    action: Required[str]

    reference_value: Required[object]

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    type_hints: Optional[object]


class ConfigJinjaNodeConfig(TypedDict, total=False):
    context_chunks_key: Optional[str]

    data_transformations: Dict[str, JinjaNodeTemplateParam]

    llm_model: Optional[str]

    log_output: bool

    log_prefix: str

    max_tokens: Optional[int]

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    output_template: JinjaNodeTemplateParam
    """
    Base model for a Jinja template. Guaranteed to store a string that can be read
    in to Template().
    """

    type_hints: Optional[object]

    verbose: bool


class ConfigProcessingNodeConfigFunctionSpecs(TypedDict, total=False):
    kwargs: Required[object]

    path: Required[str]


class ConfigProcessingNodeConfig(TypedDict, total=False):
    function_specs: Required[Dict[str, ConfigProcessingNodeConfigFunctionSpecs]]

    return_key: Required[str]

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    type_hints: Optional[object]


class ConfigRegexMatchNodeConfig(TypedDict, total=False):
    pattern: Required[str]

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    type_hints: Optional[object]


class ConfigSqlExecutorNodeConfig(TypedDict, total=False):
    connector_kwargs: Required[Dict[str, str]]

    connector_type: Literal["snowflake"]

    log_queries: bool

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    return_type: Literal["df", "dicts", "markdown", "json", "str"]

    schema_remapping_file: Optional[str]

    secrets: List[str]

    type_hints: Optional[object]


class ConfigStaticNodeConfig(TypedDict, total=False):
    from_file: Union[Iterable[object], str, object, None]

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    type_hints: Optional[object]

    value: Optional[object]


class ConfigLlmEngineNodeConfigBatchSysKwargs(TypedDict, total=False):
    checkpoint_path: Optional[str]

    labels: Optional[Dict[str, str]]

    num_shards: Optional[int]

    seed: Optional[int]


class ConfigLlmEngineNodeConfig(TypedDict, total=False):
    llm_model: Required[str]

    batch_run_mode: Literal["sync", "async"]

    batch_sys_kwargs: ConfigLlmEngineNodeConfigBatchSysKwargs

    frequency_penalty: Optional[float]

    guided_choice: Optional[List[str]]

    guided_json: Optional[object]

    guided_regex: Optional[str]

    include_stop_str_in_output: Optional[bool]

    max_tokens: Optional[int]

    node_metadata: Optional[List[str]]

    num_workers: Optional[int]

    presence_penalty: Optional[float]

    stop_sequences: Optional[List[str]]

    temperature: Optional[float]

    timeout: int

    top_k: Optional[int]

    top_p: Optional[float]

    type_hints: Optional[object]


Config: TypeAlias = Union[
    ConfigNodeConfig,
    ConfigBatchedNodeConfig,
    ConfigChunkEvaluationNodeConfig,
    ConfigRerankerNodeConfig,
    ConfigRetrieverNodeConfig,
    ConfigCitationNodeConfig,
    ConfigSearchCitationNodeConfig,
    ConfigCodeExecutionConfig,
    ConfigConditionNodeConfigInput,
    ConfigDataTransformNodeConfig,
    ConfigCreateMessagesNodeConfig,
    ConfigInsertMessagesConfig,
    ConfigRemoveMessageConfig,
    ConfigGetMessageConfig,
    ConfigTokenizerChatTemplateConfig,
    ConfigGenerationNodeConfig,
    ConfigChatGenerationNodeConfig,
    ConfigGenerationWithCitationsNodeConfig,
    ConfigChatGenerationWithCitationsNodeConfig,
    ConfigResponseParserNodeConfig,
    ConfigJinjaNodeConfig,
    ConfigProcessingNodeConfig,
    ConfigRegexMatchNodeConfig,
    ConfigSqlExecutorNodeConfig,
    ConfigStaticNodeConfig,
    ConfigLlmEngineNodeConfig,
]


class NodeItemInputParam(TypedDict, total=False):
    config: Required[Config]
    """A data model describing parameters for back-citation using ROUGE similarity.

    metric is the ROUGE metric to use (e.g. rouge1, rouge2, rougeLsum) score is one
    of "precision", "recall", "fmeasure"

    NOTE (john): copied directly from generation.py in order to subclass from
    NodeConfig.
    """

    name: Required[str]

    type: Required[str]

    inputs: Dict[str, Union[str, Dict[str, Union[str, object]]]]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable, Optional
from typing_extensions import Literal, Required, Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = ["ConfigExecuteParams", "Message", "MessageRetrievedContext", "MessageRetrievedContextContextDocument"]


class ConfigExecuteParams(TypedDict, total=False):
    session_id: Required[str]

    id: str

    concurrent: Optional[bool]

    messages: Optional[Iterable[Message]]

    metadata: Optional[object]

    return_span: Optional[bool]

    run_id: Optional[str]

    stream: bool


class MessageRetrievedContextContextDocument(TypedDict, total=False):
    document_id: Required[str]

    attachment_url: Optional[str]

    description: Optional[str]

    metadata: Optional[object]

    title: Optional[str]


class MessageRetrievedContext(TypedDict, total=False):
    id: Required[str]

    content: Required[str]

    type: Required[Literal["context_chunk", "document", "sql", "external_url", "other"]]

    context_document: Optional[MessageRetrievedContextContextDocument]
    """Represents a document related to the chat context.

    Attributes: document_id (str): The unique identifier of the document. title
    (Optional[str]): The title of the document. description (Optional[str]): The
    description of the document. attachment_url (Optional[str]): The URL of the
    original document (e.g., PDF). metadata (Optional[dict]): Additional metadata
    for the document.
    """

    page_number: Optional[str]

    score: Optional[float]

    source_url: Optional[str]


class Message(TypedDict, total=False):
    content: Required[str]

    role: Required[Literal["assistant", "user", "system"]]

    retrieved_context: Optional[Iterable[MessageRetrievedContext]]

    uuid: Annotated[str, PropertyInfo(alias="UUID")]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, List, Union, Optional
from typing_extensions import Literal, TypeAlias

from .._compat import PYDANTIC_V2
from .._models import BaseModel
from .unary_condition import UnaryCondition
from .compound_condition_output import CompoundConditionOutput

__all__ = [
    "ConfigCreateResponse",
    "Config",
    "ConfigStateMachineConfigOutput",
    "ConfigStateMachineConfigOutputMachine",
    "ConfigStateMachineConfigOutputMachineNextNode",
    "ConfigStateMachineConfigOutputMachineNextNodeCase",
    "ConfigStateMachineConfigOutputMachineNextNodeCaseCondition",
    "ConfigStateMachineConfigOutputAutoAggregatatorConfig",
    "Graph",
    "GraphEdge",
    "GraphNode",
    "GraphNodeEdge",
    "GraphNodeNode",
    "GraphNodeNodeEdge",
]

ConfigStateMachineConfigOutputMachineNextNodeCaseCondition: TypeAlias = Union[UnaryCondition, CompoundConditionOutput]


class ConfigStateMachineConfigOutputMachineNextNodeCase(BaseModel):
    condition: ConfigStateMachineConfigOutputMachineNextNodeCaseCondition
    """Representation of a boolean function with a single input e.g.

    the condition specified by input_name: x operator: 'contains' ref_value: 'c'
    would evaluate to True if x == 'cat' Operators are defined in the constant
    function store CONDITIONAL_ACTION_MAP
    """

    value: str


class ConfigStateMachineConfigOutputMachineNextNode(BaseModel):
    default: str

    cases: Optional[List[ConfigStateMachineConfigOutputMachineNextNodeCase]] = None


class ConfigStateMachineConfigOutputMachine(BaseModel):
    next_node: ConfigStateMachineConfigOutputMachineNextNode
    """A switch statement for state machines to select the next state to execute."""

    workflow_config: "WorkflowConfigOutput"

    write_to_state: Optional[Dict[str, str]] = None


class ConfigStateMachineConfigOutputAutoAggregatatorConfig(BaseModel):
    aggregate_inputs_workflow_config: Optional["WorkflowConfigOutput"] = None

    ask_inputs_workflow_config: Optional["WorkflowConfigOutput"] = None


class ConfigStateMachineConfigOutput(BaseModel):
    machine: Dict[str, ConfigStateMachineConfigOutputMachine]

    starting_node: str

    id: Optional[str] = None

    account_id: Optional[str] = None

    application_variant_id: Optional[str] = None

    auto_aggregatator_config: Optional[ConfigStateMachineConfigOutputAutoAggregatatorConfig] = None

    base_url: Optional[str] = None

    concurrency_default: Optional[bool] = None

    datasets: Optional[List[object]] = None

    done_string: Optional[str] = None

    egp_api_key_override: Optional[str] = None

    egp_ui_evaluation: Optional[object] = None

    evaluations: Optional[List["NodeItemOutput"]] = None

    final_output_nodes: Optional[List[str]] = None

    initial_state: Optional[object] = None

    nodes_to_log: Union[str, List[str], None] = None

    num_workers: Optional[int] = None

    streaming_nodes: Optional[List[str]] = None

    subtype: Optional[Literal["chat", "report"]] = None

    type: Optional[Literal["workflow", "plan", "state_machine"]] = None

    user_input: Optional[object] = None


Config: TypeAlias = Union["PlanConfigOutput", "WorkflowConfigOutput", ConfigStateMachineConfigOutput]


class GraphEdge(BaseModel):
    from_node: str

    to_node: str


class GraphNodeEdge(BaseModel):
    from_node: str

    to_node: str


class GraphNodeNodeEdge(BaseModel):
    from_node: str

    to_node: str


class GraphNodeNode(BaseModel):
    type: str

    id: Optional[str] = None

    config: Optional[object] = None

    edges: Optional[List[GraphNodeNodeEdge]] = None

    name: Optional[str] = None

    nodes: Optional[List[object]] = None


class GraphNode(BaseModel):
    type: str

    id: Optional[str] = None

    config: Optional[object] = None

    edges: Optional[List[GraphNodeEdge]] = None

    name: Optional[str] = None

    nodes: Optional[List[GraphNodeNode]] = None


class Graph(BaseModel):
    type: str

    id: Optional[str] = None

    config: Optional[object] = None

    edges: Optional[List[GraphEdge]] = None

    name: Optional[str] = None

    nodes: Optional[List[GraphNode]] = None


class ConfigCreateResponse(BaseModel):
    config: Config
    """Representation of a plan, i.e. a composition of workflows and branch complexes

    Public attributes: workflows: maps a workflow "name" to either a workflow yaml
    file or an inline workflow definition plan: representation of the graph
    connecting workflows / branch complexes

    Private attributes: helper_workflows: a list of workflows created by default, in
    order to support loops and other control flow features
    """

    graph: Graph

    user_inputs: object


from .node_item_output import NodeItemOutput
from .plan_config_output import PlanConfigOutput
from .workflow_config_output import WorkflowConfigOutput

if PYDANTIC_V2:
    ConfigCreateResponse.model_rebuild()
    ConfigStateMachineConfigOutput.model_rebuild()
    ConfigStateMachineConfigOutputMachine.model_rebuild()
    ConfigStateMachineConfigOutputMachineNextNode.model_rebuild()
    ConfigStateMachineConfigOutputMachineNextNodeCase.model_rebuild()
    ConfigStateMachineConfigOutputAutoAggregatatorConfig.model_rebuild()
    Graph.model_rebuild()
    GraphEdge.model_rebuild()
    GraphNode.model_rebuild()
    GraphNodeEdge.model_rebuild()
    GraphNodeNode.model_rebuild()
    GraphNodeNodeEdge.model_rebuild()
else:
    ConfigCreateResponse.update_forward_refs()  # type: ignore
    ConfigStateMachineConfigOutput.update_forward_refs()  # type: ignore
    ConfigStateMachineConfigOutputMachine.update_forward_refs()  # type: ignore
    ConfigStateMachineConfigOutputMachineNextNode.update_forward_refs()  # type: ignore
    ConfigStateMachineConfigOutputMachineNextNodeCase.update_forward_refs()  # type: ignore
    ConfigStateMachineConfigOutputAutoAggregatatorConfig.update_forward_refs()  # type: ignore
    Graph.update_forward_refs()  # type: ignore
    GraphEdge.update_forward_refs()  # type: ignore
    GraphNode.update_forward_refs()  # type: ignore
    GraphNodeEdge.update_forward_refs()  # type: ignore
    GraphNodeNode.update_forward_refs()  # type: ignore
    GraphNodeNodeEdge.update_forward_refs()  # type: ignore
